/*
CameraControl

CONTROLS : 
 + Arrows : move camera parallel to its frustum near plane
 + Q : move towards
 + W : move backwards
 + S : aim at the sphere
 + C : aime at the cube

*/


import processing.opengl.*;

PFont font;

PGraphics viewCamera_0;
PGraphics viewCamera_1;
damkjer.ocd.Camera cam, camSpy;

int wScene = 600;
int hScene = 600;
float aspectRatio = (float)wScene / (float) hScene;

float camRoll=0;
float camTilt=0;
float camPan=0;
float vLength = 25;
boolean drawFrustumFar = false;
boolean drawFrustumNear = true;

PVector spherePos,cubePos;
String camCommand = "";

// ------------------------------------------------
// setup
// ------------------------------------------------
void setup()
{
  font = loadFont("HelveticaNeue-Medium-12.vlw");
  
  viewCamera_0 = createGraphics(wScene,hScene,P3D);
  viewCamera_1 = createGraphics(wScene,hScene,P3D);

  camSpy     = new damkjer.ocd.Camera(viewCamera_0, aspectRatio, PI/3, 10, 200);
  cam        = new damkjer.ocd.Camera(viewCamera_1, aspectRatio, PI/3, 10, 200);


  cam.jump(0,0,50);
  cam.aim(0,0,0);

  camSpy.jump(50,-100,50);
  camSpy.aim(0,0,0);

  spherePos = new PVector(-90,-20,-100);
  cubePos = new PVector(0,0,0);


  size(2*wScene,hScene);
  smooth();
  rectMode(CENTER);
}

// ------------------------------------------------
// draw
// ------------------------------------------------
void draw()
{
  update();
  drawViewCamera_0();
  drawViewCamera_1();
  drawComposition();
}

// ------------------------------------------------
// drawViewCamera_0
// ------------------------------------------------
void drawViewCamera_0()
{
  viewCamera_0.beginDraw();
  viewCamera_0.background(255);
  viewCamera_0.translate(wScene/2,hScene/2);
  
  // Apply view camera
  camSpy.feed(viewCamera_0);
  
  // Draw scene
  drawScene(viewCamera_0);

  // Draw camera
  drawCamera(viewCamera_0,cam);

  // Get camera position in screen space
  float[] camPosition = cam.position();
  float[] camUp       = cam.up();


  // Position camera

  float x = camPosition[0]+camUp[0]*vLength;
  float y = camPosition[1]+camUp[1]*vLength;
  float z = camPosition[2]+camUp[2]*vLength;
  float camScreenX = viewCamera_0.screenX(x,y,z);
  float camScreenY = viewCamera_0.screenY(x,y,z);


  // Reset view on the buffer 
  viewCamera_0.perspective();
  viewCamera_0.camera();

  // Draw some info
  viewCamera_0.fill(0,255,0);
  viewCamera_0.textMode(SCREEN);
  viewCamera_0.textFont(font);
  viewCamera_0.text("UP",camScreenX,camScreenY);
  
  
  viewCamera_0.endDraw();
}

// ------------------------------------------------
// drawViewCamera_1
// ------------------------------------------------
void drawViewCamera_1()
{
  viewCamera_1.beginDraw();
  viewCamera_1.background(255);
  viewCamera_1.translate(wScene/2,hScene/2);

  // Apply view camera
  cam.feed(viewCamera_1);

  // Draw Scene
  drawScene(viewCamera_1);
  
  viewCamera_1.endDraw();
}

// ------------------------------------------------
// drawComposition
// ------------------------------------------------
void drawComposition()
{
  image(viewCamera_0, 0,0,wScene,hScene);
  image(viewCamera_1, wScene,0,wScene,hScene);

  // Infos
  stroke(200);
  fill(0);
  line(wScene,0,wScene,hScene);
  textFont(font);
  text("View 0",5,15);
  text("View 1"+(camCommand.equals("") ? "" : " / "+camCommand),wScene+5,15);
}

// ------------------------------------------------
// drawScene
// ------------------------------------------------
void drawScene(PGraphics g)
{
  g.pushMatrix();

  float l=35;
  g.stroke(255,0,0);g.line(0,0,0,l,0,0);
  g.stroke(0,255,0);g.line(0,0,0,0,l,0);
  g.stroke(0,0,255);g.line(0,0,0,0,0,l);
  
  g.noFill();
  g.stroke(0);
  g.pushMatrix();
  g.translate(cubePos.x,cubePos.y,cubePos.z);
  g.box(25);
  g.popMatrix();


  g.lights();
  g.noStroke();
  g.fill(209,0,0);
  g.translate(spherePos.x,spherePos.y,spherePos.z);
  g.sphereDetail(10);
  g.sphere(20);
  g.noLights();

  g.popMatrix();
  
}

// ------------------------------------------------
// drawCamera
// ------------------------------------------------
void drawCamera(PGraphics g, damkjer.ocd.Camera cam)
{
  float[] camPosition = cam.position();
  float[] camOrientation = cam.attitude();

  g.pushMatrix();

  // Position camera
  g.translate(camPosition[0], camPosition[1], camPosition[2]);
  g.rotateZ(camOrientation[2]);
  g.rotateY(camOrientation[0]);
  g.rotateX(-camOrientation[1]); // TODO : not sure of the -1 (otherwise "UP" displaying bugs)

  // Draw body of the camera
  g.lights();
  g.fill(100);
  g.pushMatrix();
  g.translate(0,0,5);
  g.box(5,5,10);
  g.popMatrix();
  g.noLights();

  // Draw axis of the camera
  g.stroke(255,0,0);g.line(0,0,0,vLength,0,0);
  g.stroke(0,255,0);g.line(0,0,0,0,vLength,0);
  g.stroke(0,0,255);g.line(0,0,0,0,0,vLength);


  // Draw frustum near + far
  g.noFill();
  g.stroke(190);
  drawFrustum(g,cam);
  g.popMatrix();
}



// ------------------------------------------------
// update
// ------------------------------------------------
void update()
{
  camCommand="";

  if (mousePressed)
  {
    float dx = ((float)mouseX - (float)pmouseX)/(float)width;
    float dy = ((float)mouseY - (float)pmouseY)/(float)height;
    camPan    = dx * PI/4;
    camTilt   = -dy * PI/5;    

    camCommand =  "roll("+nf(camRoll,0,2)+") / "+"tilt("+nf(camTilt,0,2)+")";
  }
    
  cam.tilt(camTilt);
  cam.pan(camPan);

  if (keyPressed)
  {
    if (keyCode == UP){
        cam.boom(-1);
        camCommand += (mousePressed ? " / " : "") + "boom(-1)";
    }
    if (keyCode == DOWN){
        cam.boom(1);
        camCommand += (mousePressed ? " / " : "") + "boom(+1)";
    }
    if (keyCode == LEFT){
        cam.truck(-1);
        camCommand += (mousePressed ? " / " : "") + "truck(-1)";
    }
    if (keyCode == RIGHT){
        cam.truck(1);
        camCommand += (mousePressed ? " / " : "") + "truck(+1)";
    }
    if (key == 'q' || key == 'Q'){
        cam.dolly(-1);
        camCommand += (mousePressed ? " / " : "") + "dolly(-1)";
    }
    if (key == 'w' || key == 'W'){
        cam.dolly(1);
        camCommand += (mousePressed ? " / " : "") + "dolly(+1)";
    }
    if (key == 's' || key == 'S'){
        cam.aim(spherePos.x,spherePos.y,spherePos.z);
    }
    if (key == 'c' || key == 'C'){
        cam.aim(cubePos.x,cubePos.y,cubePos.z);
    }

  }
  
}

// ------------------------------------------------
// mousePressed
// ------------------------------------------------
void mousePressed()
{
}

